# coding=utf-8
# Copyright 2024 XiaHan
# 
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.

import datetime
import os
import glob
from typing import Dict, Literal, Optional, Tuple, Union
import json
from urllib.parse import ParseResult, urljoin, urlparse
import httpx
from olah.configs import OlahConfig
from olah.constants import WORKER_API_TIMEOUT

def get_org_repo(org: Optional[str], repo: str) -> str:
    if org is None:
        org_repo = repo
    else:
        org_repo = f"{org}/{repo}"
    return org_repo

def parse_org_repo(org_repo: str) -> Tuple[Optional[str], Optional[str]]:
    if "/" in org_repo and org_repo.count("/") != 1:
        return None, None
    if "/" in org_repo:
        org, repo = org_repo.split("/")
    else:
        org = None
        repo = org_repo
    return org, repo

def get_meta_save_path(repos_path: str, repo_type: str, org: Optional[str], repo: str, commit: str) -> str:
    return os.path.join(repos_path, f"api/{repo_type}/{org}/{repo}/revision/{commit}/meta.json")

def get_meta_save_dir(repos_path: str, repo_type: str, org: Optional[str], repo: str) -> str:
    return os.path.join(repos_path, f"api/{repo_type}/{org}/{repo}/revision")

def get_file_save_path(repos_path: str, repo_type: str, org: Optional[str], repo: str, commit: str, file_path: str) -> str:
    return os.path.join(repos_path, f"heads/{repo_type}/{org}/{repo}/resolve_head/{commit}/{file_path}")

async def get_newest_commit_hf_offline(app, repo_type: Optional[Literal["models", "datasets", "spaces"]], org: str, repo: str) -> str:
    repos_path = app.app_settings.repos_path
    save_dir = get_meta_save_dir(repos_path, repo_type, org, repo)
    files = glob.glob(os.path.join(save_dir, "*", "meta.json"))

    time_revisions = []
    for file in files:
        with open(file, "r", encoding="utf-8") as f:
            obj = json.loads(f.read())
            datetime_object = datetime.datetime.fromisoformat(obj["lastModified"])
            time_revisions.append((datetime_object, obj["sha"]))

    time_revisions = sorted(time_revisions)
    return time_revisions[-1][1]

async def get_newest_commit_hf(app, repo_type: Optional[Literal["models", "datasets", "spaces"]], org: Optional[str], repo: str) -> str:
    url = urljoin(app.app_settings.config.hf_url_base(), f"/api/{repo_type}/{org}/{repo}")
    if app.app_settings.config.offline:
        return get_newest_commit_hf_offline(app, repo_type, org, repo)
    try:
        async with httpx.AsyncClient() as client:
            response = await client.get(url, timeout=WORKER_API_TIMEOUT)
            if response.status_code != 200:
                return get_newest_commit_hf_offline(app, repo_type, org, repo)
            obj = json.loads(response.text)
        return obj.get("sha", None)
    except:
        return get_newest_commit_hf_offline(app, repo_type, org, repo)

async def get_commit_hf_offline(app, repo_type: Optional[Literal["models", "datasets", "spaces"]], org: Optional[str], repo: str, commit: str) -> str:
    repos_path = app.app_settings.repos_path
    save_path = get_meta_save_path(repos_path, repo_type, org, repo, commit)
    if os.path.exists(save_path):
        with open(save_path, "r", encoding="utf-8") as f:
            obj = json.loads(f.read())
        return obj["sha"]
    else:
        return None

async def get_commit_hf(app, repo_type: Optional[Literal["models", "datasets", "spaces"]], org: Optional[str], repo: str, commit: str) -> str:
    org_repo = get_org_repo(org, repo)
    url = urljoin(app.app_settings.config.hf_url_base(), f"/api/{repo_type}/{org_repo}/revision/{commit}")
    if app.app_settings.config.offline:
        return await get_commit_hf_offline(app, repo_type, org, repo, commit)
    try:
        async with httpx.AsyncClient() as client:
            response = await client.get(url, timeout=WORKER_API_TIMEOUT, follow_redirects=True)
            if response.status_code not in [200, 307]:
                return await get_commit_hf_offline(app, repo_type, org, repo, commit)
            obj = json.loads(response.text)
        return obj.get("sha", None)
    except:
        return await get_commit_hf_offline(app, repo_type, org, repo, commit)

async def check_commit_hf(app, repo_type: Optional[Literal["models", "datasets", "spaces"]], org: Optional[str], repo: str, commit: Optional[str]=None) -> bool:
    org_repo = get_org_repo(org, repo)
    if commit is None:
        url = urljoin(app.app_settings.config.hf_url_base(), f"/api/{repo_type}/{org_repo}")
    else:
        url = urljoin(app.app_settings.config.hf_url_base(), f"/api/{repo_type}/{org_repo}/revision/{commit}")
    
    async with httpx.AsyncClient() as client:
        response = await client.get(url, timeout=WORKER_API_TIMEOUT)
    return response.status_code in [200, 307]

async def check_proxy_rules_hf(app, repo_type: Optional[Literal["models", "datasets", "spaces"]], org: Optional[str], repo: str) -> bool:
    config: OlahConfig = app.app_settings.config
    org_repo = get_org_repo(org, repo)
    return config.proxy.allow(org_repo)

async def check_cache_rules_hf(app, repo_type: Optional[Literal["models", "datasets", "spaces"]], org: Optional[str], repo: str) -> bool:
    config: OlahConfig = app.app_settings.config
    org_repo = get_org_repo(org, repo)
    return config.cache.allow(org_repo)

def get_url_tail(parsed_url: Union[str, ParseResult]) -> str:
    if isinstance(parsed_url, str):
        parsed_url = urlparse(parsed_url)
    url_tail = parsed_url.path
    if len(parsed_url.params) != 0:
        url_tail += f";{parsed_url.params}"
    if len(parsed_url.query) != 0:
        url_tail += f"?{parsed_url.query}"
    if len(parsed_url.fragment) != 0:
        url_tail += f"#{parsed_url.fragment}"
    return url_tail

def parse_range_params(file_range: str, file_size: int) -> Tuple[int, int]:
    # 'bytes=1887436800-'
    if file_range.startswith("bytes="):
        file_range = file_range[6:]
    start_pos, end_pos = file_range.split("-")
    if len(start_pos) != 0:
        start_pos = int(start_pos)
    else:
        start_pos = 0
    if len(end_pos) != 0:
        end_pos = int(end_pos)
    else:
        end_pos = file_size
    return start_pos, end_pos


class RemoteInfo(object):
    def __init__(self, method: str, url: str, headers: Dict[str, str]) -> None:
        self.method = method
        self.url = url
        self.headers = headers


# coding=utf-8
# Copyright 2024 XiaHan
# 
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.

class Bitset:
    def __init__(self, size):
        self.size = size
        self.bits = bytearray((0, ) * ((size + 7) // 8))

    def set(self, index):
        if index < 0 or index >= self.size:
            raise IndexError("Index out of range")
        byte_index = index // 8
        bit_index = index % 8
        self.bits[byte_index] |= (1 << bit_index)

    def clear(self, index):
        if index < 0 or index >= self.size:
            raise IndexError("Index out of range")
        self._resize_if_needed(index)
        byte_index = index // 8
        bit_index = index % 8
        self.bits[byte_index] &= ~(1 << bit_index)

    def test(self, index):
        if index < 0 or index >= self.size:
            raise IndexError("Index out of range")
        byte_index = index // 8
        bit_index = index % 8
        return bool(self.bits[byte_index] & (1 << bit_index))

    def __str__(self):
        return ''.join(bin(byte)[2:].zfill(8) for byte in self.bits)



# coding=utf-8
# Copyright 2024 XiaHan
# 
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.
import hashlib
import io
import os
import re
from typing import Any, Dict, List, Union
import gitdb
from git import Commit, Optional, Repo, Tree
from gitdb.base import OStream
import yaml

from olah.mirror.meta import RepoMeta
class LocalMirrorRepo(object):
    def __init__(self, path: str, repo_type: str, org: str, repo: str) -> None:
        self._path = path
        self._repo_type = repo_type
        self._org = org
        self._repo = repo

        self._git_repo = Repo(self._path)
    
    def _sha256(self, text: Union[str, bytes]) -> str:
        if isinstance(text, bytes) or isinstance(text, bytearray):
            bin = text
        elif isinstance(text, str):
            bin = text.encode('utf-8')
        else:
            raise Exception("Invalid sha256 param type.")
        sha256_hash = hashlib.sha256()
        sha256_hash.update(bin)
        hashed_string = sha256_hash.hexdigest()
        return hashed_string
    
    def _match_card(self, readme: str) -> str:
        pattern = r'\s*---(.*?)---'

        match = re.match(pattern, readme, flags=re.S)

        if match:
            card_string = match.group(1)
            return card_string
        else:
            return ""
    def _remove_card(self, readme: str) -> str:
        pattern = r'\s*---(.*?)---'
        out = re.sub(pattern, "", readme, flags=re.S)
        return out
    
    def _get_readme(self, commit: Commit) -> str:
        if "README.md" not in commit.tree:
            return ""
        else:
            out: bytes = commit.tree["README.md"].data_stream.read()
            return out.decode()
    
    def _get_description(self, commit: Commit) -> str:
        readme = self._get_readme(commit)
        return self._remove_card(readme)
    
    def _get_entry_files(self, tree, include_dir=False) -> List[str]:
        out_paths = []
        for entry in tree:
            if entry.type == "tree":
                out_paths.extend(self._get_entry_files(entry))
                if include_dir:
                    out_paths.append(entry.path)
            else:
                out_paths.append(entry.path)
        return out_paths

    def _get_tree_files(self, commit: Commit) -> List[str]:
        return self._get_entry_files(commit.tree)
    

    def _get_earliest_commit(self) -> Commit:
        earliest_commit = None
        earliest_commit_date = None

        for commit in self._git_repo.iter_commits():
            commit_date = commit.committed_datetime

            if earliest_commit_date is None or commit_date < earliest_commit_date:
                earliest_commit = commit
                earliest_commit_date = commit_date

        return earliest_commit

    def get_meta(self, commit_hash: str) -> Dict[str, Any]:
        try:
            commit = self._git_repo.commit(commit_hash)
        except gitdb.exc.BadName:
            return None
        meta = RepoMeta()
        
        meta._id = self._sha256(f"{self._org}/{self._repo}/{commit.hexsha}")
        meta.id = f"{self._org}/{self._repo}"
        meta.author = self._org
        meta.sha = commit.hexsha
        meta.lastModified = self._git_repo.head.commit.committed_datetime.strftime("%Y-%m-%dT%H:%M:%S.%fZ")
        meta.private = False
        meta.gated = False
        meta.disabled = False
        meta.tags = []
        meta.description = self._get_description(commit)
        meta.paperswithcode_id = None
        meta.downloads = 0
        meta.likes = 0
        meta.cardData = yaml.load(self._match_card(self._get_readme(commit)), Loader=yaml.CLoader)
        meta.siblings = [{"rfilename": p} for p in self._get_tree_files(commit)]
        meta.createdAt = self._get_earliest_commit().committed_datetime.strftime("%Y-%m-%dT%H:%M:%S.%fZ")
        return meta.to_dict()

    def _contain_path(self, path: str, tree: Tree) -> bool:
        norm_p = os.path.normpath(path).replace("\\", "/")
        parts = norm_p.split("/")
        for part in parts:
            if all([t.name != part for t in tree]):
                return False
            else:
                entry = tree[part]
                if entry.type == "tree":
                    tree = entry
                else:
                    tree = {}
        return True

    def get_file_head(self, commit_hash: str, path: str) -> Optional[Dict[str, Any]]:
        try:
            commit = self._git_repo.commit(commit_hash)
        except gitdb.exc.BadName:
            return None

        if not self._contain_path(path, commit.tree):
            return None
        else:
            header = {}
            header["content-length"] = str(commit.tree[path].data_stream.size)
            header["x-repo-commit"] = commit.hexsha
            header["etag"] = self._sha256(commit.tree[path].data_stream.read())
            return header

    def get_file(self, commit_hash: str, path: str) -> Optional[OStream]:
        try:
            commit = self._git_repo.commit(commit_hash)
        except gitdb.exc.BadName:
            return None
        
        def stream_wrapper(file_bytes: bytes):
            file_stream = io.BytesIO(file_bytes)
            while True:
                chunk = file_stream.read(4096)
                if len(chunk) == 0:
                    break
                else:
                    yield chunk
        
        if not self._contain_path(path, commit.tree):
            return None
        else:
            return stream_wrapper(commit.tree[path].data_stream.read())
        
    


# coding=utf-8
# Copyright 2024 XiaHan
# 
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.

import os
import struct
import threading
from typing import BinaryIO, Dict, Optional
from .bitset import Bitset

CURRENT_OLAH_CACHE_VERSION = 8
DEFAULT_BLOCK_MASK_MAX = 1024 * 1024
DEFAULT_BLOCK_SIZE = 16 * 1024 * 1024


class OlahCacheHeader(object):
    MAGIC_NUMBER = "OLAH".encode("ascii")
    HEADER_FIX_SIZE = 36

    def __init__(
        self,
        version: int = CURRENT_OLAH_CACHE_VERSION,
        block_size: int = DEFAULT_BLOCK_SIZE,
        file_size: int = 0,
    ) -> None:
        self._version = version
        self._block_size = block_size

        self._file_size = file_size
        self._block_number = (file_size + block_size - 1) // block_size

        self._block_mask_size = DEFAULT_BLOCK_MASK_MAX
        self._block_mask = Bitset(DEFAULT_BLOCK_MASK_MAX)
    
    @property
    def version(self) -> int:
        return self._version
    
    @property
    def block_size(self) -> int:
        return self._block_size

    @property
    def file_size(self) -> int:
        return self._file_size
    
    @property
    def block_number(self) -> int:
        return self._block_number

    @property
    def block_mask(self) -> int:
        return self._block_mask

    def get_header_size(self):
        return self.HEADER_FIX_SIZE + len(self._block_mask.bits)

    def _valid_header(self):
        if self._file_size > self._block_mask_size * self._block_size:
            raise Exception(
                f"The size of file {self._file_size} is out of the max capability of container ({self._block_mask_size} * {self._block_size})."
            )
        if self._version < CURRENT_OLAH_CACHE_VERSION:
            raise Exception(
                f"This Olah Cache file is created by older version Olah. Please remove cache files and retry."
            )
        
        if self._version > CURRENT_OLAH_CACHE_VERSION:
            raise Exception(
                f"This Olah Cache file is created by newer version Olah. Please remove cache files and retry."
            )

    @staticmethod
    def read(stream) -> "OlahCacheHeader":
        obj = OlahCacheHeader()
        magic, version, block_size, file_size, block_mask_size = struct.unpack(
            "<4sQQQQ", stream.read(OlahCacheHeader.HEADER_FIX_SIZE)
        )
        if magic != OlahCacheHeader.MAGIC_NUMBER:
            raise Exception("The file is not a valid olah cache file.")
        obj._version = version
        obj._block_size = block_size
        obj._file_size = file_size
        obj._block_number = (file_size + block_size - 1) // block_size
        obj._block_mask_size = block_mask_size
        obj._block_mask = Bitset(block_mask_size)
        obj._block_mask.bits = bytearray(stream.read((block_mask_size + 7) // 8))

        obj._valid_header()
        return obj

    def write(self, stream):
        btyes_header = struct.pack(
            "<4sQQQQ",
            self.MAGIC_NUMBER,
            self._version,
            self._block_size,
            self._file_size,
            self._block_mask_size,
        )
        btyes_out = btyes_header + self._block_mask.bits
        stream.write(btyes_out)

class OlahCache(object):
    def __init__(self, path: str, block_size: int = DEFAULT_BLOCK_SIZE) -> None:
        self.path: Optional[str] = path
        self.header: Optional[OlahCacheHeader] = None
        self.is_open: bool = False
        
        # Lock
        self._header_lock = threading.Lock()

        # Cache
        self._blocks_read_cache: Dict[int, bytes] = {}
        self._prefech_blocks: int = 16

        self.open(path, block_size=block_size)
        

    @staticmethod
    def create(path: str, block_size: int = DEFAULT_BLOCK_SIZE):
        return OlahCache(path, block_size=block_size)

    def open(self, path: str, block_size: int = DEFAULT_BLOCK_SIZE):
        if self.is_open:
            raise Exception("This file has been open.")
        if os.path.exists(path):
            with self._header_lock:
                with open(path, "rb") as f:
                    f.seek(0)
                    self.header = OlahCacheHeader.read(f)
        else:
            with self._header_lock:
                # Create new file
                with open(path, "wb") as f:
                    f.seek(0)
                    self.header = OlahCacheHeader(
                        version=CURRENT_OLAH_CACHE_VERSION,
                        block_size=block_size,
                        file_size=0,
                    )
                    self.header.write(f)

        self.is_open = True

    def close(self):
        if not self.is_open:
            raise Exception("This file has been close.")

        self._flush_header()
        self.path = None
        self.header = None

        self._blocks_read_cache.clear()
        
        self.is_open = False

    def _flush_header(self):
        with self._header_lock:
            with open(self.path, "rb+") as f:
                f.seek(0)
                self.header.write(f)

    def _get_file_size(self) -> int:
        with self._header_lock:
            file_size = self.header.file_size
        return file_size

    def _get_block_number(self) -> int:
        with self._header_lock:
            block_number = self.header.block_number
        return block_number
    
    def _get_block_size(self) -> int:
        with self._header_lock:
            block_size = self.header.block_size
        return block_size
    
    def _get_header_size(self) -> int:
        with self._header_lock:
            header_size = self.header.get_header_size()
        return header_size
    
    def _resize_header(self, block_num: int, file_size: int):
        with self._header_lock:
            self.header._block_number = block_num
            self.header._file_size = file_size
            self.header._valid_header()
    
    def _set_header_block(self, block_index: int):
        with self._header_lock:
            self.header.block_mask.set(block_index)

    def _test_header_block(self, block_index: int):
        with self._header_lock:
            result = self.header.block_mask.test(block_index)
        return result
    
    def _pad_block(self, raw_block: bytes):
        if len(raw_block) < self._get_block_size():
            block = raw_block + b"\x00" * (self._get_block_size() - len(raw_block))
        else:
            block = raw_block
        return block

    def flush(self):
        if not self.is_open:
            raise Exception("This file has been close.")
        self._flush_header()
    
    def has_block(self, block_index: int) -> bool:
        return self._test_header_block(block_index)

    def read_block(self, block_index: int) -> Optional[bytes]:
        if not self.is_open:
            raise Exception("This file has been closed.")

        if block_index >= self._get_block_number():
            raise Exception("Invalid block index.")
        
        # Check Cache
        if block_index in self._blocks_read_cache:
            return self._blocks_read_cache[block_index]
    
        if not self.has_block(block_index=block_index):
            return None

        offset = self._get_header_size() + (block_index * self._get_block_size())
        with open(self.path, "rb") as f:
            f.seek(offset)
            raw_block = f.read(self._get_block_size())
            # Prefetch blocks
            for block_offset in range(1, self._prefech_blocks + 1):
                if block_index + block_offset >= self._get_block_number():
                    break
                if not self.has_block(block_index=block_index):
                    self._blocks_read_cache[block_index + block_offset] = None
                else:
                    prefetch_raw_block = f.read(self._get_block_size())
                    self._blocks_read_cache[block_index + block_offset] = self._pad_block(prefetch_raw_block)

        block = self._pad_block(raw_block)
        return block
    
    def write_block(self, block_index: int, block_bytes: bytes) -> None:
        if not self.is_open:
            raise Exception("This file has been closed.")

        if block_index >= self._get_block_number():
            raise Exception("Invalid block index.")

        if len(block_bytes) != self._get_block_size():
            raise Exception("Block size does not match the cache's block size.")

        offset = self._get_header_size() + (block_index * self._get_block_size())
        with open(self.path, "rb+") as f:
            f.seek(offset)
            if (block_index + 1) * self._get_block_size() > self._get_file_size():
                real_block_bytes = block_bytes[:self._get_file_size() - block_index * self._get_block_size()]
            else:
                real_block_bytes = block_bytes
            f.write(real_block_bytes)
        
        self._set_header_block(block_index)
        self._flush_header()

        # Clear Cache
        if block_index in self._blocks_read_cache:
            del self._blocks_read_cache[block_index]
    
    def _resize_file_size(self, file_size: int):
        if not self.is_open:
            raise Exception("This file has been closed.")
        if file_size == self._get_file_size():
            return
        if file_size < self._get_file_size():
            raise Exception("Invalid resize file size. New file size must be greater than the current file size.")

        with open(self.path, "rb") as f:
            f.seek(0, os.SEEK_END)
            bin_size = f.tell()

        new_bin_size = self._get_header_size() + file_size
        with open(self.path, "rb+") as f:
            # Extend file size
            f.seek(0, os.SEEK_END)
            f.write(b'\x00' * (new_bin_size - bin_size))

    def resize(self, file_size: int):
        if not self.is_open:
            raise Exception("This file has been closed.")
        bs = self._get_block_size()
        new_block_num = (file_size + bs - 1) // bs
        self._resize_file_size(file_size)
        self._resize_header(new_block_num, file_size)
        self._flush_header()
         


# coding=utf-8
# Copyright 2024 XiaHan
# 
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.

WORKER_API_TIMEOUT = 15
CHUNK_SIZE = 4096
LFS_FILE_BLOCK = 64 * 1024 * 1024

DEFAULT_LOGGER_DIR = "./logs"

from huggingface_hub.constants import (
    REPO_TYPES_MAPPING,
    HUGGINGFACE_CO_URL_TEMPLATE,
    HUGGINGFACE_HEADER_X_REPO_COMMIT,
    HUGGINGFACE_HEADER_X_LINKED_ETAG,
    HUGGINGFACE_HEADER_X_LINKED_SIZE,
)






from fastapi import Response
from fastapi.responses import JSONResponse


def error_repo_not_found() -> JSONResponse:
    return JSONResponse(
        content={"error": "Repository not found"},
        headers={
            "x-error-code": "RepoNotFound",
            "x-error-message": "Repository not found",
        },
        status_code=401,
    )


def error_page_not_found() -> Response:
    return Response(
        headers={
            "x-error-code": "RepoNotFound",
            "x-error-message": "Sorry, we can't find the page you are looking for.",
        },
        status_code=404,
    )



# coding=utf-8
# Copyright 2024 XiaHan
# 
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.

import hashlib
import json
import os
from typing import Dict, Literal, Optional, Tuple
from fastapi import Request

from requests.structures import CaseInsensitiveDict
import httpx
from starlette.datastructures import URL
from urllib.parse import urlparse, urljoin

from olah.constants import (
    CHUNK_SIZE,
    WORKER_API_TIMEOUT,
    HUGGINGFACE_HEADER_X_REPO_COMMIT,
    HUGGINGFACE_HEADER_X_LINKED_ETAG,
    HUGGINGFACE_HEADER_X_LINKED_SIZE,
)
from olah.utils.olah_cache import OlahCache
from olah.utils.url_utils import RemoteInfo, check_cache_rules_hf, get_org_repo, get_url_tail, parse_range_params
from olah.utils.file_utils import make_dirs
from olah.constants import CHUNK_SIZE, LFS_FILE_BLOCK, WORKER_API_TIMEOUT

async def _write_cache_request(head_path: str, status_code: int, headers: Dict[str, str], content: bytes):
    rq = {
        "status_code": status_code,
        "headers": headers,
        "content": content.hex(),
    }
    with open(head_path, "w", encoding="utf-8") as f:
        f.write(json.dumps(rq, ensure_ascii=False))

async def _read_cache_request(head_path: str):
    with open(head_path, "r", encoding="utf-8") as f:
        rq = json.loads(f.read())
    
    rq["content"] = bytes.fromhex(rq["content"])
    return rq

async def _file_full_header(
        app,
        save_path: str,
        head_path: str,
        client: httpx.AsyncClient,
        method: str,
        url: str,
        headers: Dict[str, str],
        allow_cache: bool,
    ) -> Tuple[int, Dict[str, str], bytes]:
    assert method.lower() == "head"
    if not app.app_settings.config.offline:
        if os.path.exists(head_path):
            cache_rq = await _read_cache_request(head_path)
            response_headers_dict = {k.lower():v for k, v in cache_rq["headers"].items()}
            if "location" in response_headers_dict:
                parsed_url = urlparse(response_headers_dict["location"])
                if len(parsed_url.netloc) != 0:
                    new_loc = urljoin(app.app_settings.config.mirror_lfs_url_base(), get_url_tail(response_headers_dict["location"]))
                    response_headers_dict["location"] = new_loc
            return cache_rq["status_code"], response_headers_dict, cache_rq["content"]
        else:
            if "range" in headers:
                headers.pop("range")
            response = await client.request(
                method=method,
                url=url,
                headers=headers,
                timeout=WORKER_API_TIMEOUT,
            )
            response_headers_dict = {k.lower(): v for k, v in response.headers.items()}
            if allow_cache and method.lower() == "head":
                if response.status_code == 200:
                    await _write_cache_request(head_path, response.status_code, response_headers_dict, response.content)
                elif response.status_code >= 300 and response.status_code <= 399:
                    await _write_cache_request(head_path, response.status_code, response_headers_dict, response.content)
                    from_url = urlparse(url)
                    parsed_url = urlparse(response.headers["location"])
                    if len(parsed_url.netloc) != 0:
                        new_loc = urljoin(app.app_settings.config.mirror_lfs_url_base(), get_url_tail(response.headers["location"]))
                        response_headers_dict["location"] = new_loc
                else:
                    raise Exception(f"Unexpected HTTP status code {response.status_code}")
            return response.status_code, response_headers_dict, response.content
    else:
        if os.path.exists(head_path):
            cache_rq = await _read_cache_request(head_path)
            response_headers_dict = {k.lower():v for k, v in cache_rq["headers"].items()}
        else:
            response_headers_dict = {}
            cache_rq = {
                "status_code": 200,
                "headers": response_headers_dict,
                "content": b"",
            }

        new_headers = {}
        if "content-type" in response_headers_dict:
            new_headers["content-type"] = response_headers_dict["content-type"]
        if "content-length" in response_headers_dict:
            new_headers["content-length"] = response_headers_dict["content-length"]
        if HUGGINGFACE_HEADER_X_REPO_COMMIT.lower() in response_headers_dict:
            new_headers[HUGGINGFACE_HEADER_X_REPO_COMMIT.lower()] = response_headers_dict.get(HUGGINGFACE_HEADER_X_REPO_COMMIT.lower(), "")
        if HUGGINGFACE_HEADER_X_LINKED_ETAG.lower() in response_headers_dict:
            new_headers[HUGGINGFACE_HEADER_X_LINKED_ETAG.lower()] = response_headers_dict.get(HUGGINGFACE_HEADER_X_LINKED_ETAG.lower(), "")
        if HUGGINGFACE_HEADER_X_LINKED_SIZE.lower() in response_headers_dict:
            new_headers[HUGGINGFACE_HEADER_X_LINKED_SIZE.lower()] = response_headers_dict.get(HUGGINGFACE_HEADER_X_LINKED_SIZE.lower(), "")
        if "etag" in response_headers_dict:
            new_headers["etag"] = response_headers_dict["etag"]
        if "location" in response_headers_dict:
            new_headers["location"] = urljoin(app.app_settings.config.mirror_lfs_url_base(), get_url_tail(response_headers_dict["location"]))
        return cache_rq["status_code"], new_headers, cache_rq["content"]

async def _get_file_block_from_cache(cache_file: OlahCache, block_index: int):
    raw_block = cache_file.read_block(block_index)
    return raw_block

async def _get_file_block_from_remote(client: httpx.AsyncClient, remote_info: RemoteInfo, cache_file: OlahCache, block_index: int):
    block_start_pos = block_index * cache_file._get_block_size()
    block_end_pos = min(
        (block_index + 1) * cache_file._get_block_size(), cache_file._get_file_size()
    )
    remote_info.headers["range"] = f"bytes={block_start_pos}-{block_end_pos - 1}"
    raw_block = bytearray()
    async with client.stream(
        method=remote_info.method,
        url=remote_info.url,
        headers=remote_info.headers,
        timeout=WORKER_API_TIMEOUT,
    ) as response:
        response_content_length = int(response.headers['content-length'])
        async for raw_chunk in response.aiter_raw():
            if not raw_chunk:
                continue
            raw_block += raw_chunk
    # print(remote_info.url, remote_info.method, remote_info.headers)
    # print(block_start_pos, block_end_pos)
    if len(raw_block) != response_content_length:
        raise Exception(f"The content of the response is incomplete. Expected-{response_content_length}. Accepted-{len(raw_block)}")
    if len(raw_block) != (block_end_pos - block_start_pos):
        raise Exception(f"The block is incomplete. Expected-{block_end_pos - block_start_pos}. Accepted-{len(raw_block)}")
    if len(raw_block) < cache_file._get_block_size():
        raw_block += b"\x00" * (cache_file._get_block_size() - len(raw_block))
    return raw_block

async def _file_chunk_get(
    app,
    save_path: str,
    head_path: str,
    client: httpx.AsyncClient,
    method: str,
    url: str,
    headers: Dict[str, str],
    allow_cache: bool,
    file_size: int,
):
    # Redirect Chunks
    if os.path.exists(save_path):
        cache_file = OlahCache(save_path)
    else:
        cache_file = OlahCache.create(save_path)
        cache_file.resize(file_size=file_size)
    try:
        start_pos, end_pos = parse_range_params(headers.get("range", f"bytes={0}-{file_size}"), file_size)

        start_block = start_pos // cache_file._get_block_size()
        end_block = end_pos // cache_file._get_block_size()

        cur_pos = start_pos
        cur_block = start_block

        while cur_block <= end_block:
            block_start_pos = cur_block * cache_file._get_block_size()
            block_end_pos = min(
                (cur_block + 1) * cache_file._get_block_size(), file_size
            )
            if cache_file.has_block(cur_block):
                raw_block = await _get_file_block_from_cache(
                    cache_file, cur_block,
                )
            else:
                raw_block = await _get_file_block_from_remote(
                    client,
                    RemoteInfo(method, url, headers),
                    cache_file,
                    cur_block,
                )

            s = cur_pos - block_start_pos
            e = block_end_pos - block_start_pos
            chunk = raw_block[s:e]
            if len(chunk) != 0:
                yield bytes(chunk)
            cur_pos += len(chunk)

            if len(raw_block) != cache_file._get_block_size():
                raise Exception(f"The size of raw block {len(raw_block)} is different from blocksize {cache_file._get_block_size()}.")
            if not cache_file.has_block(cur_block) and allow_cache:
                cache_file.write_block(cur_block, raw_block)

            cur_block += 1
    finally:
        cache_file.close()

async def _file_chunk_head(
    app,
    save_path: str,
    head_path: str,
    client: httpx.AsyncClient,
    method: str,
    url: str,
    headers: Dict[str, str],
    allow_cache: bool,
    file_size: int,
):
    if not app.app_settings.config.offline:
        async with client.stream(
            method=method,
            url=url,
            headers=headers,
            timeout=WORKER_API_TIMEOUT,
        ) as response:
            async for raw_chunk in response.aiter_raw():
                if not raw_chunk:
                    continue
                yield raw_chunk
    else:
        yield b""


async def _file_realtime_stream(
    app,
    save_path: str,
    head_path: str,
    url: str,
    request: Request,
    method="GET",
    allow_cache=True,
    commit: Optional[str] = None,
):
    request_headers = {k: v for k, v in request.headers.items()}
    request_headers.pop("host")

    if urlparse(url).netloc == app.app_settings.config.mirror_netloc:
        hf_url = urljoin(app.app_settings.config.hf_lfs_url_base(), get_url_tail(url))
    else:
        hf_url = url

    async with httpx.AsyncClient() as client:
        # redirect_loc = await _get_redirected_url(client, method, url, request_headers)
        status_code, head_info, content = await _file_full_header(
            app=app,
            save_path=save_path,
            head_path=head_path,
            client=client,
            method="HEAD",
            url=hf_url,
            headers=request_headers,
            allow_cache=allow_cache,
        )
        if status_code != 200:
            yield status_code
            yield head_info
            yield content
            return

        file_size = int(head_info["content-length"])
        response_headers = {k: v for k,v in head_info.items()}
        if "range" in request_headers:
            start_pos, end_pos = parse_range_params(request_headers.get("range", f"bytes={0}-{file_size}"), file_size)
            response_headers["content-length"] = str(end_pos - start_pos)
        if commit is not None:
            response_headers[HUGGINGFACE_HEADER_X_REPO_COMMIT.lower()] = commit

        if app.app_settings.config.offline and "etag" not in response_headers:
            # Create fake headers when offline mode
            sha256_hash = hashlib.sha256()
            sha256_hash.update(hf_url.encode('utf-8'))
            content_hash = sha256_hash.hexdigest()
            response_headers["etag"] = f"\"{content_hash[:32]}-10\""
        yield 200
        yield response_headers
        if method.lower() == "get":
            async for each_chunk in _file_chunk_get(
                app=app,
                save_path=save_path,
                head_path=head_path,
                client=client,
                method=method,
                url=hf_url,
                headers=request_headers,
                allow_cache=allow_cache,
                file_size=file_size,
            ):
                yield each_chunk
        elif method.lower() == "head":
            async for each_chunk in _file_chunk_head(
                app=app,
                save_path=save_path,
                head_path=head_path,
                client=client,
                method=method,
                url=hf_url,
                headers=request_headers,
                allow_cache=allow_cache,
                file_size=0,
            ):
                yield each_chunk
        else:
            raise Exception(f"Unsupported method: {method}")

async def file_get_generator(
    app,
    repo_type: Literal["models", "datasets", "spaces"],
    org: str,
    repo: str,
    commit: str,
    file_path: str,
    method: Literal["HEAD", "GET"],
    request: Request,
):
    org_repo = get_org_repo(org, repo)
    # save
    repos_path = app.app_settings.repos_path
    head_path = os.path.join(
        repos_path, f"heads/{repo_type}/{org}/{repo}/resolve/{commit}/{file_path}"
    )
    save_path = os.path.join(
        repos_path, f"files/{repo_type}/{org}/{repo}/resolve/{commit}/{file_path}"
    )
    make_dirs(head_path)
    make_dirs(save_path)

    # use_cache = os.path.exists(head_path) and os.path.exists(save_path)
    allow_cache = await check_cache_rules_hf(app, repo_type, org, repo)

    # proxy
    if repo_type == "models":
        url = urljoin(app.app_settings.config.hf_url_base(), f"/{org_repo}/resolve/{commit}/{file_path}")
    else:
        url = urljoin(app.app_settings.config.hf_url_base(), f"/{repo_type}/{org_repo}/resolve/{commit}/{file_path}")
    return _file_realtime_stream(
        app=app,
        save_path=save_path,
        head_path=head_path,
        url=url,
        request=request,
        method=method,
        allow_cache=allow_cache,
        commit=commit,
    )

async def cdn_file_get_generator(
    app,
    repo_type: Literal["models", "datasets", "spaces"],
    org: str,
    repo: str,
    file_hash: str,
    method: Literal["HEAD", "GET"],
    request: Request,
):
    headers = {k: v for k, v in request.headers.items()}
    headers.pop("host")

    org_repo = get_org_repo(org, repo)
    # save
    repos_path = app.app_settings.repos_path
    head_path = os.path.join(
        repos_path, f"heads/{repo_type}/{org}/{repo}/cdn/{file_hash}"
    )
    save_path = os.path.join(
        repos_path, f"files/{repo_type}/{org}/{repo}/cdn/{file_hash}"
    )
    make_dirs(head_path)
    make_dirs(save_path)

    # use_cache = os.path.exists(head_path) and os.path.exists(save_path)
    allow_cache = await check_cache_rules_hf(app, repo_type, org, repo)

    # proxy
    # request_url = urlparse(str(request.url))
    # if request_url.netloc == app.app_settings.config.hf_lfs_netloc:
    #     redirected_url = urljoin(app.app_settings.config.mirror_lfs_url_base(), get_url_tail(request_url))
    # else:
    #     redirected_url = urljoin(app.app_settings.config.mirror_url_base(), get_url_tail(request_url))

    return _file_realtime_stream(
        app=app,
        save_path=save_path,
        head_path=head_path,
        url=str(request.url),
        request=request,
        method=method,
        allow_cache=allow_cache,
    )



# coding=utf-8
# Copyright 2024 XiaHan
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.

from contextlib import asynccontextmanager
import os
import argparse
import traceback
from typing import Annotated, Optional, Union
from urllib.parse import urljoin
from fastapi import FastAPI, Header, Request
from fastapi.responses import FileResponse, HTMLResponse, StreamingResponse, Response, JSONResponse
from fastapi_utils.tasks import repeat_every
import git
import httpx
from pydantic import BaseSettings
from olah.configs import OlahConfig
from olah.errors import error_repo_not_found, error_page_not_found
from olah.mirror.repos import LocalMirrorRepo
from olah.proxy.files import cdn_file_get_generator, file_get_generator
from olah.proxy.lfs import lfs_get_generator, lfs_head_generator
from olah.proxy.meta import meta_generator, meta_proxy_cache
from olah.utils.url_utils import check_proxy_rules_hf, check_commit_hf, get_commit_hf, get_newest_commit_hf, parse_org_repo
from olah.constants import REPO_TYPES_MAPPING
from olah.utils.logging import build_logger

# ======================
# Utilities
# ======================
async def check_connection(url: str) -> bool:
    try:
        async with httpx.AsyncClient() as client:
            response = await client.request(
                method="HEAD",
                url=url,
                timeout=10,
            )
        if response.status_code != 200:
            return False
        else:
            return True
    except httpx.TimeoutException:
        return False


@repeat_every(seconds=60)
async def check_hf_connection() -> None:
    if app.app_settings.config.offline:
        return
    hf_online_status = await check_connection(
        "https://huggingface.co/datasets/Salesforce/wikitext/resolve/main/.gitattributes"
    )
    if not hf_online_status:
        logger.info(
            "Cannot reach Huggingface Official Site. Trying to connect hf-mirror."
        )
        hf_mirror_online_status = await check_connection(
            "https://hf-mirror.com/datasets/Salesforce/wikitext/resolve/main/.gitattributes"
        )
        if not hf_online_status and not hf_mirror_online_status:
            logger.error("Failed to reach Huggingface Official Site.")
            logger.error("Failed to reach hf-mirror Site.")


@asynccontextmanager
async def lifespan(app: FastAPI):
    await check_hf_connection()
    yield

# ======================
# Application
# ======================
app = FastAPI(lifespan=lifespan, debug=False)

class AppSettings(BaseSettings):
    # The address of the model controller.
    config: OlahConfig = OlahConfig()
    repos_path: str = "./repos"

# ======================
# API Hooks
# ======================
async def meta_proxy_common(repo_type: str, org: str, repo: str, commit: str, request: Request) -> Response:
    if repo_type not in REPO_TYPES_MAPPING.keys():
        return error_page_not_found()
    if not await check_proxy_rules_hf(app, repo_type, org, repo):
        return error_repo_not_found()
    # Check Mirror Path
    for mirror_path in app.app_settings.config.mirrors_path:
        try:
            git_path = os.path.join(mirror_path, repo_type, org, repo)
            if os.path.exists(git_path):
                local_repo = LocalMirrorRepo(git_path, repo_type, org, repo)
                meta_data = local_repo.get_meta(commit)
                if meta_data is None:
                    continue
                return JSONResponse(content=meta_data)
        except git.exc.InvalidGitRepositoryError:
            logger.warning(f"Local repository {git_path} is not a valid git reposity.")
            continue

    # Proxy the HF File Meta
    try:
        if not app.app_settings.config.offline and not await check_commit_hf(
            app, repo_type, org, repo, commit
        ):
            return error_repo_not_found()
        commit_sha = await get_commit_hf(app, repo_type, org, repo, commit)
        if commit_sha is None:
            return error_repo_not_found()
        # if branch name and online mode, refresh branch info
        if not app.app_settings.config.offline and commit_sha != commit:
            await meta_proxy_cache(app, repo_type, org, repo, commit, request)

        generator = meta_generator(app, repo_type, org, repo, commit_sha, request)
        headers = await generator.__anext__()
        return StreamingResponse(generator, headers=headers)
    except httpx.ConnectTimeout:
        traceback.print_exc()
        return Response(status_code=504)


@app.get("/api/{repo_type}/{org_repo}")
async def meta_proxy(repo_type: str, org_repo: str, request: Request):
    org, repo = parse_org_repo(org_repo)
    if org is None and repo is None:
        return error_repo_not_found()
    if not app.app_settings.config.offline:
        new_commit = await get_newest_commit_hf(app, repo_type, org, repo)
    else:
        new_commit = "main"
    return await meta_proxy_common(
        repo_type=repo_type, org=org, repo=repo, commit=new_commit, request=request
    )


@app.get("/api/{repo_type}/{org}/{repo}/revision/{commit}")
async def meta_proxy_commit2(
    repo_type: str, org: str, repo: str, commit: str, request: Request
):
    return await meta_proxy_common(
        repo_type=repo_type, org=org, repo=repo, commit=commit, request=request
    )


@app.get("/api/{repo_type}/{org_repo}/revision/{commit}")
async def meta_proxy_commit(repo_type: str, org_repo: str, commit: str, request: Request):
    org, repo = parse_org_repo(org_repo)
    if org is None and repo is None:
        return error_repo_not_found()

    return await meta_proxy_common(
        repo_type=repo_type, org=org, repo=repo, commit=commit, request=request
    )


# ======================
# File Head Hooks
# ======================
async def file_head_common(
    repo_type: str, org: str, repo: str, commit: str, file_path: str, request: Request
) -> Response:
    if repo_type not in REPO_TYPES_MAPPING.keys():
        return error_page_not_found()
    if not await check_proxy_rules_hf(app, repo_type, org, repo):
        return error_repo_not_found()

    # Check Mirror Path
    for mirror_path in app.app_settings.config.mirrors_path:
        try:
            git_path = os.path.join(mirror_path, repo_type, org, repo)
            if os.path.exists(git_path):
                local_repo = LocalMirrorRepo(git_path, repo_type, org, repo)
                head = local_repo.get_file_head(commit_hash=commit, path=file_path)
                if head is None:
                    continue
                return Response(headers=head)
        except git.exc.InvalidGitRepositoryError:
            logger.warning(f"Local repository {git_path} is not a valid git reposity.")
            continue
    
    # Proxy the HF File Head
    try:
        if not app.app_settings.config.offline and not await check_commit_hf(
            app, repo_type, org, repo, commit
        ):
            return error_repo_not_found()
        commit_sha = await get_commit_hf(app, repo_type, org, repo, commit)
        if commit_sha is None:
            return error_repo_not_found()
        generator = await file_get_generator(
            app,
            repo_type,
            org,
            repo,
            commit_sha,
            file_path=file_path,
            method="HEAD",
            request=request,
        )
        status_code = await generator.__anext__()
        headers = await generator.__anext__()
        return StreamingResponse(generator, headers=headers, status_code=status_code)
    except httpx.ConnectTimeout:
        traceback.print_exc()
        return Response(status_code=504)


@app.head("/{repo_type}/{org}/{repo}/resolve/{commit}/{file_path:path}")
async def file_head3(
    repo_type: str, org: str, repo: str, commit: str, file_path: str, request: Request
):
    return await file_head_common(
        repo_type=repo_type,
        org=org,
        repo=repo,
        commit=commit,
        file_path=file_path,
        request=request,
    )


@app.head("/{org_or_repo_type}/{repo_name}/resolve/{commit}/{file_path:path}")
async def file_head2(
    org_or_repo_type: str, repo_name: str, commit: str, file_path: str, request: Request
):
    if org_or_repo_type in ["models", "datasets", "spaces"]:
        repo_type: str = org_or_repo_type
        org, repo = parse_org_repo(repo_name)
        if org is None and repo is None:
            return error_repo_not_found()
    else:
        repo_type: str = "models"
        org, repo = org_or_repo_type, repo_name

    return await file_head_common(
        repo_type=repo_type,
        org=org,
        repo=repo,
        commit=commit,
        file_path=file_path,
        request=request,
    )


@app.head("/{org_repo}/resolve/{commit}/{file_path:path}")
async def file_head(org_repo: str, commit: str, file_path: str, request: Request):
    repo_type: str = "models"
    org, repo = parse_org_repo(org_repo)
    if org is None and repo is None:
        return error_repo_not_found()
    return await file_head_common(
        repo_type=repo_type,
        org=org,
        repo=repo,
        commit=commit,
        file_path=file_path,
        request=request,
    )


@app.head("/{org_repo}/{hash_file}")
@app.head("/{repo_type}/{org_repo}/{hash_file}")
async def cdn_file_head(org_repo: str, hash_file: str, request: Request, repo_type: str = "models"):
    org, repo = parse_org_repo(org_repo)
    if org is None and repo is None:
        return error_repo_not_found()

    if not await check_proxy_rules_hf(app, repo_type, org, repo):
        return error_repo_not_found()

    try:
        generator = await cdn_file_get_generator(app, repo_type, org, repo, hash_file, method="HEAD", request=request)
        status_code = await generator.__anext__()
        headers = await generator.__anext__()
        return StreamingResponse(generator, headers=headers, status_code=status_code)
    except httpx.ConnectTimeout:
        return Response(status_code=504)


# ======================
# File Hooks
# ======================
async def file_get_common(
    repo_type: str, org: str, repo: str, commit: str, file_path: str, request: Request
) -> Response:
    if repo_type not in REPO_TYPES_MAPPING.keys():
        return error_page_not_found()
    if not await check_proxy_rules_hf(app, repo_type, org, repo):
        return error_repo_not_found()
    # Check Mirror Path
    for mirror_path in app.app_settings.config.mirrors_path:
        try:
            git_path = os.path.join(mirror_path, repo_type, org, repo)
            if os.path.exists(git_path):
                local_repo = LocalMirrorRepo(git_path, repo_type, org, repo)
                content_stream = local_repo.get_file(commit_hash=commit, path=file_path)
                if content_stream is None:
                    continue
                return StreamingResponse(content_stream)
        except git.exc.InvalidGitRepositoryError:
            logger.warning(f"Local repository {git_path} is not a valid git reposity.")
            continue
    try:
        if not app.app_settings.config.offline and not await check_commit_hf(app, repo_type, org, repo, commit):
            return error_repo_not_found()
        commit_sha = await get_commit_hf(app, repo_type, org, repo, commit)
        if commit_sha is None:
            return error_repo_not_found()
        generator = await file_get_generator(
            app,
            repo_type,
            org,
            repo,
            commit_sha,
            file_path=file_path,
            method="GET",
            request=request,
        )
        status_code = await generator.__anext__()
        headers = await generator.__anext__()
        return StreamingResponse(generator, headers=headers, status_code=status_code)
    except httpx.ConnectTimeout:
        traceback.print_exc()
        return Response(status_code=504)


@app.get("/{repo_type}/{org}/{repo}/resolve/{commit}/{file_path:path}")
async def file_get3(org: str, repo: str, commit: str, file_path: str, request: Request, repo_type: str):
    return await file_get_common(
        repo_type=repo_type,
        org=org,
        repo=repo,
        commit=commit,
        file_path=file_path,
        request=request,
    )

@app.get("/{org_or_repo_type}/{repo_name}/resolve/{commit}/{file_path:path}")
async def file_get2(org_or_repo_type: str, repo_name: str, commit: str, file_path: str, request: Request):
    if org_or_repo_type in ["models", "datasets", "spaces"]:
        repo_type: str = org_or_repo_type
        org, repo = parse_org_repo(repo_name)
        if org is None and repo is None:
            return error_repo_not_found()
    else:
        repo_type: str = "models"
        org, repo = org_or_repo_type, repo_name

    return await file_get_common(
        repo_type=repo_type,
        org=org,
        repo=repo,
        commit=commit,
        file_path=file_path,
        request=request,
    )

@app.get("/{org_repo}/resolve/{commit}/{file_path:path}")
async def file_get(org_repo: str, commit: str, file_path: str, request: Request):
    repo_type: str = "models"
    org, repo = parse_org_repo(org_repo)
    if org is None and repo is None:
        return error_repo_not_found()

    return await file_get_common(
        repo_type=repo_type,
        org=org,
        repo=repo,
        commit=commit,
        file_path=file_path,
        request=request,
    )

@app.get("/{org_repo}/{hash_file}")
@app.get("/{repo_type}/{org_repo}/{hash_file}")
async def cdn_file_get(org_repo: str, hash_file: str, request: Request, repo_type: str = "models"):
    org, repo = parse_org_repo(org_repo)
    if org is None and repo is None:
        return error_repo_not_found()

    if not await check_proxy_rules_hf(app, repo_type, org, repo):
        return error_repo_not_found()
    try:
        generator = await cdn_file_get_generator(app, repo_type, org, repo, hash_file, method="GET", request=request)
        status_code = await generator.__anext__()
        headers = await generator.__anext__()
        return StreamingResponse(generator, headers=headers, status_code=status_code)
    except httpx.ConnectTimeout:
        return Response(status_code=504)

# ======================
# LFS Hooks
# ======================
@app.head("/repos/{dir1}/{dir2}/{hash_repo}/{hash_file}")
async def lfs_head(dir1: str, dir2: str, hash_repo: str, hash_file: str, request: Request):
    try:
        generator = await lfs_head_generator(app, dir1, dir2, hash_repo, hash_file, request)
        status_code = await generator.__anext__()
        headers = await generator.__anext__()
        return StreamingResponse(generator, headers=headers, status_code=status_code)
    except httpx.ConnectTimeout:
        return Response(status_code=504)

@app.get("/repos/{dir1}/{dir2}/{hash_repo}/{hash_file}")
async def lfs_get(dir1: str, dir2: str, hash_repo: str, hash_file: str, request: Request):
    try:
        generator = await lfs_get_generator(app, dir1, dir2, hash_repo, hash_file, request)
        status_code = await generator.__anext__()
        headers = await generator.__anext__()
        return StreamingResponse(generator, headers=headers, status_code=status_code)
    except httpx.ConnectTimeout:
        return Response(status_code=504)

# ======================
# Web Page Hooks
# ======================
@app.get("/", response_class=HTMLResponse)
async def index():
    with open(os.path.join(os.path.dirname(__file__), "../static/index.html"), "r", encoding="utf-8") as f:
        page = f.read()
    return page

if __name__ in ["__main__", "olah.server"]:
    parser = argparse.ArgumentParser(
        description="Olah Huggingface Mirror Server."
    )
    parser.add_argument("--config", "-c", type=str, default="")
    parser.add_argument("--host", type=str, default="localhost")
    parser.add_argument("--port", type=int, default=8090)
    parser.add_argument("--ssl-key", type=str, default=None, help="The SSL key file path, if HTTPS is used")
    parser.add_argument("--ssl-cert", type=str, default=None, help="The SSL cert file path, if HTTPS is used")
    parser.add_argument("--repos-path", type=str, default="./repos", help="The folder to save cached repositories")
    parser.add_argument("--log-path", type=str, default="./logs", help="The folder to save logs")
    args = parser.parse_args()
    print(args)

    logger = build_logger("olah", "olah.log", logger_dir=args.log_path)
    
    def is_default_value(args, arg_name):
        if hasattr(args, arg_name):
            arg_value = getattr(args, arg_name)
            arg_default = parser.get_default(arg_name)
            return arg_value == arg_default
        return False

    if args.config != "":
        config = OlahConfig(args.config)
    else:
        config = OlahConfig()
    
    if is_default_value(args, "host"):
        args.host = config.host
    if is_default_value(args, "port"):
        args.port = config.port
    if is_default_value(args, "ssl_key"):
        args.ssl_key = config.ssl_key
    if is_default_value(args, "ssl_cert"):
        args.ssl_cert = config.ssl_cert
    if is_default_value(args, "repos_path"):
        args.repos_path = config.repos_path

    app.app_settings = AppSettings(
        config=config,
        repos_path=args.repos_path,
    )

    import uvicorn
    if __name__ == "__main__":
        uvicorn.run(
            "olah.server:app",
            host=args.host,
            port=args.port,
            log_level="info",
            reload=False,
            ssl_keyfile=args.ssl_key,
            ssl_certfile=args.ssl_cert
        )



# coding=utf-8
# Copyright 2024 XiaHan
# 
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.

import os
import shutil
import tempfile
from typing import Dict, Literal
from urllib.parse import urljoin
from fastapi import FastAPI, Request

import httpx
from olah.constants import CHUNK_SIZE, WORKER_API_TIMEOUT

from olah.utils.url_utils import check_cache_rules_hf, get_org_repo
from olah.utils.file_utils import make_dirs


async def meta_cache_generator(app: FastAPI, save_path: str):
    yield {}
    with open(save_path, "rb") as f:
        while True:
            chunk = f.read(CHUNK_SIZE)
            if not chunk:
                break
            yield chunk

async def meta_proxy_cache(
    app: FastAPI,
    repo_type: Literal["models", "datasets", "spaces"],
    org: str,
    repo: str,
    commit: str,
    request: Request,
):
    # save
    repos_path = app.app_settings.repos_path
    save_dir = os.path.join(
        repos_path, f"api/{repo_type}/{org}/{repo}/revision/{commit}"
    )
    save_path = os.path.join(save_dir, "meta.json")
    make_dirs(save_path)

    # url
    org_repo = get_org_repo(org, repo)
    meta_url = urljoin(
        app.app_settings.config.hf_url_base(),
        f"/api/{repo_type}/{org_repo}/revision/{commit}",
    )
    async with httpx.AsyncClient() as client:
        response = await client.request(
            method="GET",
            url=meta_url,
            timeout=WORKER_API_TIMEOUT,
            follow_redirects=True,
        )
        if response.status_code == 200:
            with open(save_path, "wb") as meta_file:
                meta_file.write(response.content)
        else:
            raise Exception(f"Cannot get the branch info from the url {meta_url}, status: {response.status_code}")

async def meta_proxy_generator(
    app: FastAPI,
    headers: Dict[str, str],
    meta_url: str,
    allow_cache: bool,
    save_path: str,
):
    try:
        temp_file_path = None
        async with httpx.AsyncClient(follow_redirects=True) as client:
            with tempfile.NamedTemporaryFile(mode="wb", delete=True) as temp_file:
                temp_file_path = temp_file.name
                if not allow_cache:
                    write_temp_file = False
                else:
                    write_temp_file = True
                async with client.stream(
                    method="GET",
                    url=meta_url,
                    headers=headers,
                    timeout=WORKER_API_TIMEOUT,
                ) as response:
                    response_headers = response.headers
                    yield response_headers

                    async for raw_chunk in response.aiter_raw():
                        if not raw_chunk:
                            continue
                        if write_temp_file:
                            temp_file.write(raw_chunk)
                        yield raw_chunk
                if temp_file_path is not None:
                    temp_file.flush()
                    shutil.copyfile(temp_file_path, save_path)
    finally:
        if temp_file_path is not None and os.path.exists(temp_file_path):
            os.remove(temp_file_path)


async def meta_generator(
    app: FastAPI,
    repo_type: Literal["models", "datasets", "spaces"],
    org: str,
    repo: str,
    commit: str,
    request: Request,
):
    headers = {k: v for k, v in request.headers.items()}
    headers.pop("host")

    # save
    repos_path = app.app_settings.repos_path
    save_dir = os.path.join(
        repos_path, f"api/{repo_type}/{org}/{repo}/revision/{commit}"
    )
    save_path = os.path.join(save_dir, "meta.json")
    make_dirs(save_path)

    use_cache = os.path.exists(save_path)
    allow_cache = await check_cache_rules_hf(app, repo_type, org, repo)

    org_repo = get_org_repo(org, repo)
    meta_url = urljoin(
        app.app_settings.config.hf_url_base(),
        f"/api/{repo_type}/{org_repo}/revision/{commit}",
    )
    # proxy
    if use_cache:
        async for item in meta_cache_generator(app, save_path):
            yield item
    else:
        async for item in meta_proxy_generator(
            app, headers, meta_url, allow_cache, save_path
        ):
            yield item
